#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>
#include <unistd.h>
#include <errno.h>
#include "requestpool.h"

//#define DEBUG_REQUESTPOOL

#ifdef DEBUG_REQUESTPOOL
  #define LP_DEBUG(msg...)     do {          printf(msg); fflush(stdout); }   while (0)
  #define LP_DEBUGT(t, msg...) do { if (t) { printf(msg); fflush(stdout); } } while (0)
#else
  #define LP_DEBUG(msg...)     do { } while (0)
  #define LP_DEBUGT(t, msg...) do { } while (0)
#endif

RequestPool::RequestPool()
{
	//start with an invalid pool
	requestPool.quit = 0;
	requestPool.valid = 0;
	requestPool.destroyed = 1;
	requestCount = 0;
	requestPool.head = requestPool.tail = NULL;
}

int RequestPool::Initialize( int aStackSize )
{
	LP_DEBUG("RequestPool: Initialize(%d).\n", aStackSize);

	int r;

	if (aStackSize == 0 || aStackSize < DEFAULT_STACK_SIZE)
	{
		aStackSize = DEFAULT_STACK_SIZE;
	}

	while(!requestPool.destroyed)
	{
		//wait for async destroy
		LP_DEBUG("RequestPool: Initialize(). Waiting for the pool to be destroyed!\n");
		usleep(MUTEX_WAIT_DESTROY_TIME);
	}

	r = pthread_mutex_init(&requestPool.mutex, NULL);
	assert(0 == r);

	r = pthread_cond_init(&requestPool.cond, NULL);
	assert(0 == r);

	r = pthread_attr_init(&handlerThreadAttr);
	assert(0 == r);

	r = pthread_attr_setdetachstate(&handlerThreadAttr, PTHREAD_CREATE_DETACHED);
	assert(0 == r);

	r = pthread_attr_setstacksize(&handlerThreadAttr, aStackSize);
	assert(0 == r);


	requestPool.head = requestPool.tail = NULL;
	requestPool.quit = 0;
	requestPool.valid = 1;
	requestPool.destroyed = 0;
	requestCount = 0;
	pendingArgument = NULL;

	r = pthread_create( &handlerThread, &handlerThreadAttr, &RequestPool::ThreadFunction, this);
	LP_DEBUG("RequestPool: Initialize returned %d.\n", r);

	return r;

}

int RequestPool::Add( void (*function)(void *),  void * argument, int direction )
{
	LP_DEBUG("RequestPool: Add(0x%x, 0x%x).\n", function, argument);

	int r;
	request_item_t* request;

	request = (request_item_t*)malloc(sizeof(request_item_t));

	if(request)
	{
		request->function = function;
		request->argument = argument;

		r = pthread_mutex_lock(&requestPool.mutex);

		if (r != EINVAL)
       		{
			if (requestPool.valid)
			{
				if (direction == RP_ADD_HEAD)
				{
					request->next = requestPool.head;
					requestPool.head = request;
				}
				else
				{
					request->next = NULL;
					if (requestPool.head == NULL)
					{
						requestPool.head =  request;
					}
					else
					{
						requestPool.tail->next = request;
					}
					requestPool.tail = request;
				}
	
				requestCount+=1;

				r = pthread_cond_signal(&requestPool.cond);
				assert(0 == r);
	
			}
			else
			{
				r = EINVAL;
			}

			r = pthread_mutex_unlock(&requestPool.mutex);
			assert(0 == r);
		}
		else
		{
			assert(false); //todo: for a while
		}
	}
	else
	{
		r = EAGAIN;
	}
	LP_DEBUG("RequestPool: Add returned %d.\n", r );
	return r;
}

int RequestPool::Add2( void (*function1)(void *),  void * argument1, void (*function2)(void *), void * argument2, int direction )
{
	LP_DEBUG("RequestPool: Add2(0x%x, 0x%x).\n", function1, argument1, function2, argument2);

	int r;
	request_item_t* request1;
	request_item_t* request2;

	request1 = (request_item_t*)malloc(sizeof(request_item_t));
	request2 = (request_item_t*)malloc(sizeof(request_item_t));

	if ( request1 && request2 )
	{
		request1->function = function1;
		request1->argument = argument1;
		request2->function = function2;
		request2->argument = argument2;
		request1->next = request2;

		r = pthread_mutex_lock(&requestPool.mutex);

		if (r != EINVAL)
       		{
			if (requestPool.valid)
			{
				if (direction == RP_ADD_HEAD)
				{
					request2->next = requestPool.head;
					requestPool.head = request1;
				}
				else
				{
					request2->next = NULL;
					if (requestPool.head == NULL)
					{
						requestPool.head =  request1;
					}
					else
					{
						requestPool.tail->next = request1;
					}
					requestPool.tail = request2;
				}
	
				requestCount+=2;

				r = pthread_cond_signal(&requestPool.cond);
				assert(0 == r);
	
			}
			else
			{
				r = EINVAL;
			}

			r = pthread_mutex_unlock(&requestPool.mutex);
			assert(0 == r);
		}
		else
		{
			assert(false); //todo: for a while
		}
	}
	else
	{
		r = EAGAIN;
	}
	LP_DEBUG("RequestPool: Add2 returned %d.\n", r );
	return r;
}


void RequestPool::Empty( void )
{

	int r;

	LP_DEBUG("RequestPool: Empty().\n");

	r = pthread_mutex_lock(&requestPool.mutex);
	assert(0 == r);

	while ( requestPool.head )
	{
		request_item_t*	request = requestPool.head;

		requestPool.head = request->next;
		if (requestPool.tail == request)
		{
			requestPool.tail =  NULL;
		}
		free(request->argument);
		free(request);
	}

	requestCount = 0;

	r = pthread_mutex_unlock(&requestPool.mutex);
	assert(0 == r);

	LP_DEBUG("RequestPool: Empty.\n");
}

int RequestPool::Count(void)
{
	return requestCount;
}

void RequestPool::FreePool()
{
	LP_DEBUG("RequestPool: FreePool().\n");

	int r;

	//clean up the pool
	while ( requestPool.head )
	{
		request_item_t*	request = requestPool.head;

		requestPool.head = request->next;
		if (requestPool.tail == request)
		{
			requestPool.tail =  NULL;
		}
		free(request->argument);
		free(request);
	}
	if ( pendingArgument )
		free( pendingArgument );

	requestCount = 0;
	requestPool.quit = 1;
	requestPool.valid = 0;

	r = pthread_mutex_unlock(&requestPool.mutex);
	assert(0 == r);

	r = pthread_mutex_destroy(&requestPool.mutex);
	while(r == EBUSY)
	{
		//somebody still trying to do something on the pool, wait till it understands that it is impossible (via valid)
		LP_DEBUG("RequestPool: FreePool(). Waiting on busy mutex!\n");
		usleep(MUTEX_WAIT_RELEASE_TIME);
		r = pthread_mutex_destroy(&requestPool.mutex);
	} 
	r = pthread_cond_destroy(&requestPool.cond);
	assert(0 == r);

	r = pthread_attr_destroy(&handlerThreadAttr);
	assert(0 == r);

	requestPool.destroyed = 1;

	LP_DEBUG("RequestPool: FreePool returned.\n");
}

void RequestPool::ThreadFreePool(void* aPtr)
{

	assert( aPtr );

	RequestPool* pThis = static_cast<RequestPool*> (aPtr);

	pThis->FreePool();

}

int RequestPool::Destroy( int aDestructionType )
{
	LP_DEBUG("RequestPool: Destroy(%d).\n", aDestructionType);
	int	r;

	if(aDestructionType == RP_DESTROY_NOW)
	{
		r = pthread_mutex_lock(&requestPool.mutex);
		assert( 0 == r );

		r = pthread_cancel(handlerThread);
		assert( 0 == r );

		r = pthread_mutex_unlock(&requestPool.mutex);
		assert(0 == r);
	}
	else if ( aDestructionType == RP_DESTROY_ASYNC)
	{
		Add(NULL, NULL, RP_ADD_TAIL);
	}
	else if ( aDestructionType == RP_DESTROY_SYNC)
	{
		r = pthread_mutex_lock(&requestPool.mutex);
		assert( 0 == r );
		if( r != EINVAL)
		{
			if (requestPool.valid)
			{

				requestPool.quit = 1;
				requestPool.valid = 0;

				r = pthread_cond_signal(&requestPool.cond);
				assert(0 == r);

				r = pthread_cond_wait(&requestPool.cond, &requestPool.mutex);
				assert(0 == r);

				r = pthread_mutex_unlock(&requestPool.mutex);
				assert(0 == r);

				r = pthread_mutex_destroy(&requestPool.mutex);
				while(r == EBUSY)
				{
					LP_DEBUG("RequestPool: Destroy(). Waiting on busy mutex!\n");
					usleep(MUTEX_WAIT_RELEASE_TIME); //wait for the mutex to be released
					r = pthread_mutex_destroy(&requestPool.mutex);
				} 
			
				r = pthread_cond_destroy(&requestPool.cond);
				assert(0 == r);
				r = pthread_attr_destroy(&handlerThreadAttr);
				assert(0 == r);

				requestPool.destroyed = 1;
			}
			else
			{
				//let FreePool do the business
				r = pthread_mutex_unlock(&requestPool.mutex);
				assert(0 == r);
			}
		}
       	}
	LP_DEBUG("RequestPool: Destroy returned.\n");
	return r;
}

void RequestPool::RequestHandler()
{
	LP_DEBUG("RequestPool: RequestHandler().\n");

	int		r;
	request_item_t*	request;

	r = pthread_mutex_lock(&requestPool.mutex);
	assert(0 == r);

    	pthread_cleanup_push(&RequestPool::ThreadFreePool, this);

	while (1)
	{
		if ( !requestPool.head && !requestPool.tail )
		{
			LP_DEBUG("RequestPool: waiting for a job...\n");
			r = pthread_cond_wait(&requestPool.cond, &requestPool.mutex);
			assert(0 == r);
		        LP_DEBUG("RequestPool: job received!\n");
		}

		if ( request = requestPool.head )
		{
		    void (* function)(void *) = request->function;
		    void  * argument          = request->argument;
		    pendingArgument = argument;

		    if( function == NULL )
		    {
		        LP_DEBUG("RequestPool: canclelation job! Exiting...\n");
			pthread_exit(0); //will call clean up handler
		    }

		    // Pop the item from the queue.
		    requestPool.head = request->next;

		    if (requestPool.tail == request)
		    {
			requestPool.tail =  NULL;
		    }
		    free(request);
		    requestCount-=1;

		    r = pthread_mutex_unlock(&requestPool.mutex);
		    assert(0 == r);

		    pthread_cleanup_push((void(*)(void*)) pthread_mutex_lock, &requestPool.mutex);
		    function(argument);
		    pendingArgument = NULL;
		    free(argument);
		    pthread_cleanup_pop(1);
		   
		}
		if ( !requestPool.head && !requestPool.tail && requestPool.quit )
		{
			break;
		}
	}

	r = pthread_cond_signal(&requestPool.cond);
	assert(0 == r);

	r = pthread_mutex_unlock(&requestPool.mutex);
	assert(0 == r);

	pthread_cleanup_pop(0);

	LP_DEBUG("RequestPool: RequestHandler exited.\n");
}

void* RequestPool::ThreadFunction(void* aPtr)
{

	assert( aPtr );

	RequestPool* pThis = static_cast<RequestPool*> (aPtr);

	pThis->RequestHandler();

	return NULL;
}

